#!/bin/bash

# Copyright (c) 2020 by Robert Bosch GmbH. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
SCRIPT_DIR=`dirname $(realpath -s $0)`

echo $SCRIPT_DIR

build_iceoryx_docker() {
	if [[ "$(docker images -q iceoryx:latest 2> /dev/null)" == "" ]]; then
		echo "Building base iceoryx docker image."
		$SCRIPT_DIR/../../tools/docker/build_iceoryx_docker.sh
	fi
}

echo "Select an example:"
echo ""
options=("Simple Internode-Communcation" "Quit")
select opt in "${options[@]}"
do
    case $opt in
        "Simple Internode-Communcation")
			build_iceoryx_docker
			cd $SCRIPT_DIR/docker
			docker-compose -f simple_internode_communication.yml up
            break
            ;;
        "Quit")
            break
            ;;
        *) echo invalid option;;

    esac
done
